class Renderer(object):
    def __init__(self, processor_options=None):
        # processors is defined on the base class to provide a common way for users to
        # add to and manipulate them before calling render()
        self.processors = self.default_processors()
        self.processor_options = processor_options or {}

    def default_processors(self):
        ''' 
        Return a list of processors that this renderer uses by default
        '''
        raise NotImplementedError()

    def preprocess(self, root_frame):
        frame = root_frame
        for processor in self.processors:
            frame = processor(frame, options=self.processor_options)
        return frame

    def render(self, session):
        '''
        Return a string that contains the rendered form of `frame`
        '''
        raise NotImplementedError()



# this script just does a busywait for 0.25 seconds.
from __future__ import print_function
import time, sys

def do_nothing():
    pass

def busy_wait(duration):
    end_time = time.time() + duration

    while time.time() < end_time:
        do_nothing()

def main():
    print('sys.argv: ', sys.argv)
    busy_wait(0.25)


if __name__ == '__main__':
    main()



import sys, os, codecs, runpy, tempfile, glob, time, fnmatch, optparse
import pyinstrument
from pyinstrument import Profiler, renderers
from pyinstrument.session import ProfilerSession
from pyinstrument.util import object_with_import_path
from pyinstrument.vendor.six import exec_, PY2


def main():
    usage = ("usage: pyinstrument [options] scriptfile [arg] ...")
    version_string = 'pyinstrument {v}, on Python {pyv[0]}.{pyv[1]}.{pyv[2]}'.format(
        v=pyinstrument.__version__,
        pyv=sys.version_info,
    )
    parser = optparse.OptionParser(usage=usage, version=version_string)
    parser.allow_interspersed_args = False

    def dash_m_callback(option, opt, value, parser):
        parser.values.module_name = value
        # everything after the -m argument should be passed to that module
        parser.values.module_args = parser.rargs + parser.largs
        parser.rargs[:] = []
        parser.largs[:] = []

    parser.add_option('', '--load-prev',
        dest='load_prev', action='store', metavar='ID',
        help="Instead of running a script, load a previous report")

    parser.add_option('-m', '',
        dest='module_name', action='callback', callback=dash_m_callback,
        type="str",
        help="run library module as a script, like 'python -m module'")

    parser.add_option('-o', '--outfile',
        dest="outfile", action='store',
        help="save to <outfile>", default=None)

    parser.add_option('-r', '--renderer',
        dest='renderer', action='store', type='string',
        help=("how the report should be rendered. One of: 'text', 'html', 'json', or python "
              "import path to a renderer class"),
        default='text')

    parser.add_option('', '--html',
        dest="output_html", action='store_true',
        help=optparse.SUPPRESS_HELP, default=False)  # deprecated shortcut for --renderer=html

    parser.add_option('-t', '--timeline',
        dest='timeline', action='store_true',
        help="render as a timeline - preserve ordering and don't condense repeated calls")

    parser.add_option('', '--hide',
        dest='hide_fnmatch', action='store', metavar='EXPR',
        help=("glob-style pattern matching the file paths whose frames to hide. Defaults to "
              "'*{sep}lib{sep}*'.").format(sep=os.sep),
        default='*{sep}lib{sep}*'.format(sep=os.sep))
    parser.add_option('', '--hide-regex',
        dest='hide_regex', action='store', metavar='REGEX',
        help=("regex matching the file paths whose frames to hide. Useful if --hide doesn't give "
              "enough control."))

    parser.add_option('', '--show',
        dest='show_fnmatch', action='store', metavar='EXPR',
        help=("glob-style pattern matching the file paths whose frames to "
              "show, regardless of --hide or --hide-regex. For example, use "
              "--show '*/<library>/*' to show frames within a library that "
              "would otherwise be hidden."))
    parser.add_option('', '--show-regex',
        dest='show_regex', action='store', metavar='REGEX',
        help=("regex matching the file paths whose frames to always show. "
              "Useful if --show doesn't give enough control."))
    parser.add_option('', '--show-all',
        dest='show_all', action='store_true',
        help="show everything", default=False)

    parser.add_option('', '--unicode',
        dest='unicode', action='store_true',
        help='(text renderer only) force unicode text output')
    parser.add_option('', '--no-unicode',
        dest='unicode', action='store_false',
        help='(text renderer only) force ascii text output')

    parser.add_option('', '--color',
        dest='color', action='store_true',
        help='(text renderer only) force ansi color text output')
    parser.add_option('', '--no-color',
        dest='color', action='store_false',
        help='(text renderer only) force no color text output')

    if not sys.argv[1:]:
        parser.print_help()
        sys.exit(2)

    options, args = parser.parse_args()

    if args == [] and options.module_name is None and options.load_prev is None:
        parser.print_help()
        sys.exit(2)

    if not options.hide_regex:
        options.hide_regex = fnmatch.translate(options.hide_fnmatch)
    
    if not options.show_regex and options.show_fnmatch:
        options.show_regex = fnmatch.translate(options.show_fnmatch)
     
    if options.show_all:
        options.show_regex = r'.*'

    if options.load_prev:
        session = load_report(options.load_prev)
    else:
        if options.module_name is not None:
            sys.argv[:] = [options.module_name] + options.module_args
            code = "run_module(modname, run_name='__main__')"
            globs = {
                'run_module': runpy.run_module,
                'modname': options.module_name
            }
        else:
            sys.argv[:] = args
            progname = args[0]
            sys.path.insert(0, os.path.dirname(progname))
            with open(progname, 'rb') as fp:
                code = compile(fp.read(), progname, 'exec')
            globs = {
                '__file__': progname,
                '__name__': '__main__',
                '__package__': None,
            }

        profiler = Profiler()

        profiler.start()

        try:
            exec_(code, globs, None)
        except (SystemExit, KeyboardInterrupt):
            pass

        profiler.stop()
        session = profiler.last_session

    if options.output_html:
        options.renderer = 'html'

    output_to_temp_file = (options.renderer == 'html'
                           and not options.outfile
                           and file_is_a_tty(sys.stdout))

    if options.outfile:
        f = codecs.open(options.outfile, 'w', 'utf-8')
        should_close_f_after_writing = True
    elif not output_to_temp_file:
        if PY2:
            f = codecs.getwriter('utf-8')(sys.stdout)
        else:
            f = sys.stdout
        should_close_f_after_writing = False

    renderer_kwargs = {'processor_options': {
        'hide_regex': options.hide_regex,
        'show_regex': options.show_regex,
    }}

    if options.timeline is not None:
        renderer_kwargs['timeline'] = options.timeline

    if options.renderer == 'text':
        unicode_override = options.unicode != None
        color_override = options.color != None
        unicode = options.unicode if unicode_override else file_supports_unicode(f)
        color = options.color if color_override else file_supports_color(f)
        
        renderer_kwargs.update({'unicode': unicode, 'color': color})

    renderer_class = get_renderer_class(options.renderer)
    renderer = renderer_class(**renderer_kwargs)

    # remove this frame from the trace
    renderer.processors.append(remove_first_pyinstrument_frame_processor)


    if output_to_temp_file:
        output_filename = renderer.open_in_browser(session)
        print('stdout is a terminal, so saved profile output to %s' % output_filename)
    else:
        f.write(renderer.render(session))
        if should_close_f_after_writing:
            f.close()

    if options.renderer == 'text':
        _, report_identifier = save_report(session)
        print('To view this report with different options, run:')
        print('    pyinstrument --load-prev %s [options]' % report_identifier)
        print('')


def file_supports_color(file_obj):
    """
    Returns True if the running system's terminal supports color.

    Borrowed from Django
    https://github.com/django/django/blob/master/django/core/management/color.py
    """
    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (plat != 'win32' or
                                                  'ANSICON' in os.environ)

    is_a_tty = file_is_a_tty(file_obj)

    return (supported_platform and is_a_tty)


def file_supports_unicode(file_obj):
    encoding = getattr(file_obj, 'encoding', None)
    if not encoding:
        return False

    codec_info = codecs.lookup(encoding)

    return ('utf' in codec_info.name)


def file_is_a_tty(file_obj):
    return hasattr(file_obj, 'isatty') and file_obj.isatty()


def get_renderer_class(renderer):
    if renderer == 'text':
        return renderers.ConsoleRenderer
    elif renderer == 'html':
        return renderers.HTMLRenderer
    elif renderer == 'json':
        return renderers.JSONRenderer
    else:
        return object_with_import_path(renderer)


def report_dir():
    report_dir = os.path.join(tempfile.gettempdir(), 'pyinstrument')
    if not os.path.exists(report_dir):
        os.mkdir(report_dir)
    return report_dir

def load_report(identifier=None):
    '''
    Returns the session referred to by identifier
    '''
    path = os.path.join(
        report_dir(),
        identifier + '.pyireport'
    )
    return ProfilerSession.load(path)

def save_report(session):
    '''
    Saves the session to a temp file, and returns that path.
    Also prunes the number of reports to 10 so there aren't loads building up.
    '''
    # prune this folder to contain the last 10 sessions
    previous_reports = glob.glob(os.path.join(report_dir(), '*.pyireport'))
    previous_reports.sort(reverse=True)
    while len(previous_reports) > 10:
        report_file = previous_reports.pop()
        os.remove(report_file)

    identifier = time.strftime('%Y-%m-%dT%H-%M-%S', time.localtime(session.start_time))

    path = os.path.join(
        report_dir(),
        identifier + '.pyireport'
    )
    session.save(path)
    return path, identifier

# pylint: disable=W0613
def remove_first_pyinstrument_frame_processor(frame, options):
    '''
    The first frame when using the command line is always the __main__ function. I want to remove
    that from the output.
    '''
    if frame is None:
        return None

    if 'pyinstrument' in frame.file_path and len(frame.children) == 1:
        frame = frame.children[0]
        frame.remove_from_parent()
        return frame

    return frame


if __name__ == '__main__':
    main()



import os, io
import codecs
import tempfile
import webbrowser
from pyinstrument.renderers.base import Renderer
from pyinstrument.renderers.jsonrenderer import JSONRenderer
from pyinstrument import processors


class HTMLRenderer(Renderer):
    def render(self, session):
        resources_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'html_resources/')

        if not os.path.exists(os.path.join(resources_dir, 'app.js')):
            raise RuntimeError("Could not find app.js. If you are running "
                               "pyinstrument from a git checkout, run 'python "
                               "setup.py build' to compile the Javascript "
                               "(requires nodejs).")

        with io.open(os.path.join(resources_dir, 'app.js'), encoding='utf-8') as f:
            js = f.read()

        session_json = self.render_json(session)

        page = u'''<!DOCTYPE html>
            <html>
            <head>
                <meta charset="utf-8">
            </head>
            <body>
                <div id="app"></div>
                <script>
                    window.profileSession = {session_json}
                </script>
                <script>
                    {js}
                </script>
            </body>
            </html>'''.format(js=js, session_json=session_json)

        return page

    def open_in_browser(self, session, output_filename=None):
        """
        Open the rendered HTML in a webbrowser.

        If output_filename=None (the default), a tempfile is used.

        The filename of the HTML file is returned.

        """
        if output_filename is None:
            output_file = tempfile.NamedTemporaryFile(suffix='.html', delete=False)
            output_filename = output_file.name
            with codecs.getwriter('utf-8')(output_file) as f:
                f.write(self.render(session))
        else:
            with codecs.open(output_filename, 'w', 'utf-8') as f:
                f.write(self.render(session))

        from pyinstrument.vendor.six.moves import urllib
        url = urllib.parse.urlunparse(('file', '', output_filename, '', '', ''))
        webbrowser.open(url)
        return output_filename

    def render_json(self, session):
        json_renderer = JSONRenderer()
        json_renderer.processors = self.processors
        return json_renderer.render(session)

    def default_processors(self):
        return [
            processors.remove_importlib,
            processors.merge_consecutive_self_time,
            processors.aggregate_repeated_calls,
            processors.group_library_frames_processor,
            processors.remove_unnecessary_self_time_nodes,
            processors.remove_irrelevant_nodes,
        ]



import importlib, warnings
from pyinstrument.vendor.decorator import decorator

def object_with_import_path(import_path):
    if '.' not in import_path:
        raise ValueError("Can't import '%s', it is not a valid import path" % import_path)
    module_path, object_name = import_path.rsplit('.', 1)

    module = importlib.import_module(module_path)
    return getattr(module, object_name)

def truncate(string, max_length):
    if len(string) > max_length:
        return string[0:max_length-3]+'...'
    return string

@decorator
def deprecated(func, *args, **kwargs):
    ''' Marks a function as deprecated. '''
    warnings.warn(
        '{} is deprecated and should no longer be used.'.format(func),
        DeprecationWarning,
        stacklevel=3
    )
    return func(*args, **kwargs)

def deprecated_option(option_name, message=''):
    ''' Marks an option as deprecated. '''
    def caller(func, *args, **kwargs):
        if option_name in kwargs:
            warnings.warn(
                '{} is deprecated. {}'.format(option_name, message),
                DeprecationWarning,
                stacklevel=3
            )
            
        return func(*args, **kwargs)
    return decorator(caller)



# coding: utf8
import time
import pyinstrument
from pyinstrument.renderers.base import Renderer
from pyinstrument import processors
from pyinstrument.util import truncate


class ConsoleRenderer(Renderer):
    def __init__(self, unicode=False, color=False, show_all=False, timeline=False, **kwargs):
        super(ConsoleRenderer, self).__init__(**kwargs)

        self.unicode = unicode
        self.color = color
        self.colors = self.colors_enabled if color else self.colors_disabled
        if show_all:
            self.processors.remove(processors.group_library_frames_processor)
        if timeline:
            self.processors.remove(processors.aggregate_repeated_calls)

    def render(self, session):
        result = self.render_preamble(session)

        self.root_frame = self.preprocess(session.root_frame())

        if self.root_frame is None:
            result += 'No samples were recorded.\n\n'
            return result

        result += self.render_frame(self.root_frame)
        result += '\n'

        return result

    # pylint: disable=W1401
    def render_preamble(self, session):
        lines = [
            r"",
            r"  _     ._   __/__   _ _  _  _ _/_  ",
            r" /_//_/// /_\ / //_// / //_'/ //    ",
            r"/   _/        {:>20}".format('v'+pyinstrument.__version__),
        ]

        lines[1] += " Recorded: {:<9}".format(time.strftime('%X', time.localtime(session.start_time)))
        lines[2] += " Duration: {:<9.3f}".format(session.duration)
        lines[1] += " Samples:  {}".format(session.sample_count)
        if session.cpu_time is not None:
            lines[2] += " CPU time: {:.3f}".format(session.cpu_time)

        lines.append('')
        lines.append('Program: %s' % session.program)
        lines.append('')
        lines.append('')

        return '\n'.join(lines)

    def render_frame(self, frame, indent=u'', child_indent=u''):
        if not frame.group or (frame.group.root == frame
                               or frame.total_self_time > 0.2*self.root_frame.time()
                               or frame in frame.group.exit_frames):
            time_str = (self._ansi_color_for_time(frame)
                        + '{:.3f}'.format(frame.time()) 
                        + self.colors.end)
            function_color = self._ansi_color_for_function(frame)
            result = u'{indent}{time_str} {function_color}{function}{c.end}  {c.faint}{code_position}{c.end}\n'.format(
                indent=indent,
                time_str=time_str,
                function_color=function_color,
                function=frame.function,
                code_position=frame.code_position_short,
                c=self.colors)
            if self.unicode:
                indents = {'├': u'├─ ', '│': u'│  ', '└': u'└─ ', ' ': u'   '}
            else:
                indents = {'├': u'|- ', '│': u'|  ', '└': u'`- ', ' ': u'   '}

            if frame.group and frame.group.root == frame:
                result += u'{indent}[{count} frames hidden]  {c.faint}{libraries}{c.end}\n'.format(
                    indent=child_indent+u'   ',
                    count=len(frame.group.frames),
                    libraries=truncate(', '.join(frame.group.libraries), 40),
                    c=self.colors)
                for key in indents:
                    indents[key] = u'      '
        else:
            result = ''
            indents = {'├': u'', '│': u'', '└': u'', ' ': u''}

        if frame.children:
            last_child = frame.children[-1]

        for child in frame.children:
            if child is not last_child:
                c_indent = child_indent + indents['├']
                cc_indent = child_indent + indents['│']
            else:
                c_indent = child_indent + indents['└']
                cc_indent = child_indent + indents[' ']
            result += self.render_frame(child, indent=c_indent, child_indent=cc_indent)

        return result

    def _ansi_color_for_time(self, frame):
        proportion_of_total = frame.time() / self.root_frame.time()
        
        if proportion_of_total > 0.6:
            return self.colors.red
        elif proportion_of_total > 0.2:
            return self.colors.yellow
        elif proportion_of_total > 0.05:
            return self.colors.green
        else:
            return self.colors.bright_green + self.colors.faint

    def _ansi_color_for_function(self, frame):
        if frame.is_application_code:
            return self.colors.bg_dark_blue_255+self.colors.white_255
        else:
            return ''

    def default_processors(self):
        return [
            processors.remove_importlib,
            processors.merge_consecutive_self_time,
            processors.aggregate_repeated_calls,
            processors.group_library_frames_processor,
            processors.remove_unnecessary_self_time_nodes,
            processors.remove_irrelevant_nodes,
        ]
     
    class colors_enabled:
        red = '\033[31m'
        green = '\033[32m'
        yellow = '\033[33m'
        blue = '\033[34m'
        cyan = '\033[36m'
        bright_green = '\033[92m'
        white = '\033[37m\033[97m'

        bg_dark_blue_255 = '\033[48;5;24m'
        white_255 = '\033[38;5;15m'

        bold = '\033[1m'
        faint = '\033[2m'

        end = '\033[0m'


    class colors_disabled:
        def __getattr__(self, key):
            return ''

    colors_disabled = colors_disabled()



import io, json
from collections import deque
from pyinstrument.frame import Frame, SelfTimeFrame
from pyinstrument.vendor.six import PY2

ASSERTION_MESSAGE = ('Please raise an issue at http://github.com/pyinstrument/issues and '
                     'let me know how you caused this error!')

class ProfilerSession(object):
    def __init__(self, frame_records, start_time, duration, sample_count, start_call_stack, 
                 program, cpu_time=None):
        self.frame_records = frame_records
        self.start_time = start_time
        self.duration = duration
        self.sample_count = sample_count
        self.start_call_stack = start_call_stack
        self.program = program
        self.cpu_time = cpu_time

    @staticmethod
    def load(filename):
        with io.open(filename, 'rb' if PY2 else 'r') as f:
            return ProfilerSession.from_json(json.load(f))

    def save(self, filename):
        with io.open(filename, 'wb' if PY2 else 'w') as f:
            json.dump(self.to_json(), f)

    def to_json(self):
        return {
            'frame_records': self.frame_records,
            'start_time': self.start_time,
            'duration': self.duration,
            'sample_count': self.sample_count,
            'start_call_stack': self.start_call_stack,
            'program': self.program,
            'cpu_time': self.cpu_time,
        }

    @staticmethod
    def from_json(json_dict):
        return ProfilerSession(
            frame_records=json_dict['frame_records'],
            start_time=json_dict['start_time'],
            duration=json_dict['duration'],
            sample_count=json_dict['sample_count'],
            start_call_stack=json_dict['start_call_stack'],
            program=json_dict['program'],
            cpu_time=json_dict['cpu_time'],
        )

    def root_frame(self, trim_stem=True):
        ''' 
        Parses the internal frame records and returns a tree of Frame objects
        '''
        root_frame = None

        frame_stack = []

        for frame_tuple in self.frame_records:
            identifier_stack = frame_tuple[0]
            time = frame_tuple[1]

            # now we must create a stack of frame objects and assign this time to the leaf
            for stack_depth, frame_identifier in enumerate(identifier_stack):
                if stack_depth < len(frame_stack):
                    if frame_identifier != frame_stack[stack_depth].identifier:
                        # trim any frames after and including this one
                        del frame_stack[stack_depth:]

                if stack_depth >= len(frame_stack):
                    frame = Frame(frame_identifier)
                    frame_stack.append(frame)
                    
                    if stack_depth == 0:
                        # There should only be one root frame, as far as I know
                        assert root_frame is None, ASSERTION_MESSAGE
                        root_frame = frame
                    else:
                        parent = frame_stack[stack_depth-1]
                        parent.add_child(frame)

            # trim any extra frames
            del frame_stack[stack_depth+1:]  # pylint: disable=W0631

            # assign the time to the final frame
            frame_stack[-1].add_child(SelfTimeFrame(self_time=time))
        
        if root_frame is None:
            return None
        
        if trim_stem:
            root_frame = self._trim_stem(root_frame)
        
        return root_frame

    def _trim_stem(self, frame):
        # trim the start of the tree before any branches.
        # we also don't want to trim beyond the call to profiler.start()

        start_stack = deque(self.start_call_stack)
        if start_stack.popleft() != frame.identifier:
            # the frame doesn't match where the profiler was started. Don't trim.
            return frame

        while frame.self_time == 0 and len(frame.children) == 1:
            # check child matches the start_call_stack, otherwise stop descending
            if len(start_stack) == 0 or frame.children[0].identifier != start_stack.popleft():
                break

            frame = frame.children[0]

        frame.remove_from_parent()
        return frame



# this script just does a busywait for 0.25 seconds.
import time

def do_nothing():
    pass

def busy_wait(duration):
    end_time = time.time() + duration

    while time.time() < end_time:
        do_nothing()

if __name__ == '__main__':
    busy_wait(0.25)



from __future__ import print_function
import time, json

from pyinstrument import Profiler, renderers

# Utilities #

def do_nothing():
    pass

def busy_wait(duration):
    end_time = time.time() + duration

    while time.time() < end_time:
        do_nothing()

def long_function_a():
    time.sleep(0.25)

def long_function_b():
    time.sleep(0.5)

# Tests #

def test_collapses_multiple_calls_by_default():
    profiler = Profiler()
    profiler.start()

    long_function_a()
    long_function_b()
    long_function_a()
    long_function_b()

    profiler.stop()

    text_output = profiler.output_text()

    # output should be something like:
    # 1.513 test_collapses_multiple_calls_by_default  test/test_profiler.py:25
    # |- 0.507 long_function_a  test/test_profiler.py:17
    # |- 0.503 long_function_b  test/test_profiler.py:20

    assert text_output.count('test_collapses_multiple_calls_by_default') == 1
    assert text_output.count('long_function_a') == 1
    assert text_output.count('long_function_b') == 1

def test_profiler_retains_multiple_calls():
    profiler = Profiler()
    profiler.start()

    long_function_a()
    long_function_b()
    long_function_a()
    long_function_b()

    profiler.stop()

    print(profiler.output_text())

    frame = profiler.last_session.root_frame()
    assert frame.function == 'test_profiler_retains_multiple_calls'
    assert len(frame.children) == 4

def test_two_functions():
    profiler = Profiler()
    profiler.start()

    long_function_a()
    long_function_b()

    profiler.stop()

    print(profiler.output_text())

    frame = profiler.last_session.root_frame()

    assert frame.function == 'test_two_functions'
    assert len(frame.children) == 2

    frame_b, frame_a = sorted(frame.children, key=lambda f: f.time(), reverse=True)

    assert frame_a.function == 'long_function_a'
    assert frame_b.function == 'long_function_b'
    assert 0.2 < frame_a.time() < 0.3
    assert 0.45 < frame_b.time() < 0.55

def test_context_manager():
    with Profiler() as profiler:
        long_function_a()
        long_function_b()

    frame = profiler.last_session.root_frame()
    assert frame.function == 'test_context_manager'
    assert len(frame.children) == 2

def test_json_output():
    with Profiler() as profiler:
        long_function_a()
        long_function_b()

    output_data = profiler.output(renderers.JSONRenderer(), root=True)

    output = json.loads(output_data)
    assert 'root_frame' in output

    root_frame = output['root_frame']

    assert root_frame['function'] == 'test_json_output'
    assert len(root_frame['children']) == 2

def test_empty_profile():
    with Profiler() as profiler:
        pass
    profiler.output(renderer=renderers.ConsoleRenderer())



import sys, os, uuid


class BaseFrame(object):
    def __init__(self, parent=None, self_time=0):
        self.parent = parent
        self._self_time = self_time
        self.group = None

    # pylint: disable=W0212
    def remove_from_parent(self):
        '''
        Removes this frame from its parent, and nulls the parent link
        '''
        if self.parent:
            self.parent._children.remove(self)
            self.parent._invalidate_time_caches()
            self.parent = None

    @property
    def proportion_of_parent(self):
        if self.parent:
            try:
                return self.time() / self.parent.time()
            except ZeroDivisionError:
                return float('nan')
        else:
            return 1.0

    @property
    def total_self_time(self):
        '''
        The total amount of self time in this frame (including self time recorded by SelfTimeFrame
        children)
        '''
        self_time = self.self_time
        for child in self.children:
            if isinstance(child, SelfTimeFrame):
                self_time += child.self_time
        return self_time

    @property
    def self_time(self):
        return self._self_time

    @self_time.setter
    def self_time(self, self_time):
        self._self_time = self_time
        self._invalidate_time_caches()

    # invalidates the cache for the time() function.
    # called whenever self_time or _children is modified.
    def _invalidate_time_caches(self):
        pass

    # stylistically I'd rather this was a property, but using @property appears to use twice
    # as many stack frames, so I'm forced into using a function since this method is recursive
    # down the call tree.
    def time(self): raise NotImplementedError()

    @property
    def function(self): raise NotImplementedError()

    @property
    def file_path(self): raise NotImplementedError()

    @property
    def line_no(self): raise NotImplementedError()

    @property
    def file_path_short(self): raise NotImplementedError()

    @property
    def is_application_code(self): raise NotImplementedError()

    @property
    def code_position_short(self): raise NotImplementedError()

    @property
    def children(self): raise NotImplementedError()


class Frame(BaseFrame):
    """
    Object that represents a stack frame in the parsed tree
    """
    def __init__(self, identifier='', parent=None, children=None, self_time=0):
        super(Frame, self).__init__(parent=parent, self_time=self_time)

        self.identifier = identifier
        self._children = []

        self._time = None

        if children:
            for child in children:
                self.add_child(child)
    
    def add_child(self, frame, after=None):
        '''
        Adds a child frame, updating the parent link.
        Optionally, insert the frame in a specific position by passing the frame to insert
        this one after.
        '''
        frame.remove_from_parent()
        frame.parent = self
        if after is None:
            self._children.append(frame)
        else:
            index = self._children.index(after) + 1
            self._children.insert(index, frame)

        self._invalidate_time_caches()
    
    def add_children(self, frames, after=None):
        '''
        Convenience method to add multiple frames at once.
        '''
        if after is not None:
            # if there's an 'after' parameter, add the frames in reverse so the order is
            # preserved.
            for frame in reversed(frames):
                self.add_child(frame, after=after)
        else:
            for frame in frames:
                self.add_child(frame)

    @property
    def children(self):
        # Return an immutable copy (this property should only be mutated using methods)
        # Also, returning a copy avoid problems when mutating while iterating, which happens a lot
        # in processors!
        return tuple(self._children)

    @property
    def function(self):
        if self.identifier:
            return self.identifier.split('\x00')[0]

    @property
    def file_path(self):
        if self.identifier:
            return self.identifier.split('\x00')[1]

    @property
    def line_no(self):
        if self.identifier:
            return int(self.identifier.split('\x00')[2])

    @property
    def file_path_short(self):
        """ Return the path resolved against the closest entry in sys.path """
        if not hasattr(self, '_file_path_short'):
            if self.file_path:
                result = None

                for path in sys.path:
                    # On Windows, if self.file_path and path are on different drives, relpath
                    # will result in exception, because it cannot compute a relpath in this case.
                    # The root cause is that on Windows, there is no root dir like '/' on Linux.
                    try:
                        candidate = os.path.relpath(self.file_path, path)
                    except ValueError:
                        continue

                    if not result or (len(candidate.split(os.sep)) < len(result.split(os.sep))):
                        result = candidate

                self._file_path_short = result
            else:
                self._file_path_short = None

        return self._file_path_short

    @property
    def is_application_code(self):
        if self.identifier:
            return (('%slib%s' % (os.sep, os.sep)) not in self.file_path
                    and '<frozen importlib._bootstrap' not in self.file_path)

    @property
    def code_position_short(self):
        if self.identifier:
            return '%s:%i' % (self.file_path_short, self.line_no)

    def time(self):
        if self._time is None:
            # can't use a sum(<generator>) expression here sadly, because this method
            # recurses down the call tree, and the generator uses an extra stack frame,
            # meaning we hit the stack limit when the profiled code is 500 frames deep.
            self._time = self.self_time

            for child in self.children:
                self._time += child.time()

        return self._time

    # pylint: disable=W0212
    def _invalidate_time_caches(self):
        self._time = None
        # null all the parent's caches also.
        frame = self
        while frame.parent is not None:
            frame = frame.parent
            frame._time = None


    def __repr__(self):
        return 'Frame(identifier=%s, time=%f, len(children)=%d), group=%r' % (
            self.identifier, self.time(), len(self.children), self.group
        )


class SelfTimeFrame(BaseFrame):
    """
    Represents a time spent inside a function
    """
    def time(self): 
        return self.self_time
    
    @property
    def function(self): return '[self]'

    @property
    def _children(self): return []

    @property
    def children(self): return []

    @property
    def file_path(self): return self.parent.file_path

    @property
    def line_no(self): return self.parent.line_no

    @property
    def file_path_short(self): return ''

    @property
    def is_application_code(self): return False

    @property
    def code_position_short(self): return ''

    @property
    def identifier(self): return '[self]'


class FrameGroup(object):
    def __init__(self, root, **kwargs):
        super(FrameGroup, self).__init__(**kwargs)
        self.root = root
        self.id = str(uuid.uuid4())
        self._frames = []
        self._exit_frames = None
        self._libraries = None

        self.add_frame(root)
    
    @property
    def libraries(self):
        if self._libraries is None:
            libraries = []
            for frame in self.frames:
                library = frame.file_path_short.split(os.sep)[0]
                library, _ = os.path.splitext(library)
                if library and library not in libraries:
                    libraries.append(library)
            self._libraries = libraries      

        return self._libraries

    @property
    def frames(self):
        return tuple(self._frames)

    # pylint: disable=W0212
    def add_frame(self, frame):
        if frame.group:
            frame.group._frames.remove(frame)

        self._frames.append(frame)
        frame.group = self

    @property
    def exit_frames(self):
        '''
        Returns a list of frames whose children include a frame outside of the group
        '''
        if self._exit_frames is None:
            exit_frames = []
            for frame in self.frames:
                if any(c.group != self for c in frame.children):
                    exit_frames.append(frame)
            self._exit_frames = exit_frames

        return self._exit_frames

    def __repr__(self):
        return 'FrameGroup(len(frames)=%d)' % len(self.frames)



from __future__ import print_function
import sys, subprocess, os


def test_command_line():
    test_script = os.path.join(os.path.dirname(__file__), 'busywait.py')
    output = subprocess.check_output([sys.executable, '-m', 'pyinstrument', test_script])

    assert 'busy_wait' in str(output)
    assert 'do_nothing' in str(output)

def test_module_running():
    working_dir = os.path.dirname(os.path.dirname(__file__))
    output = subprocess.check_output(
        [sys.executable, '-m', 'pyinstrument', '-m', 'test.busywait_module'],
        cwd=working_dir
    )

    assert 'busy_wait' in str(output)
    assert 'do_nothing' in str(output)

def test_module_args():
    working_dir = os.path.dirname(os.path.dirname(__file__))
    output = subprocess.check_output(
        [sys.executable, '-m', 'pyinstrument', '-m', 'test.busywait_module', '--test-argument'],
        cwd=working_dir
    )

    assert "sys.argv:  ['test.busywait_module', '--test-argument']" in str(output)

def test_running_yourself_as_module():
    working_dir = os.path.dirname(os.path.dirname(__file__))
    subprocess.check_call(
        [sys.executable, '-m', 'pyinstrument', '-m', 'pyinstrument'],
        cwd=working_dir
    )




# -*- coding: utf-8 -*-
import timeit, time, sys, inspect
from pyinstrument import renderers
from pyinstrument.session import ProfilerSession
from pyinstrument.util import deprecated, deprecated_option
from pyinstrument_cext import setstatprofile

try:
    from time import process_time
except ImportError:
    process_time = None

timer = timeit.default_timer




class Profiler(object):
    # pylint: disable=W0613
    @deprecated_option('use_signal')
    @deprecated_option('recorder')
    def __init__(self, interval=0.001, use_signal=None, recorder=None):
        self.interval = interval
        self.last_profile_time = 0.0
        self.frame_records = []
        self._start_time = None
        self._start_process_time = None
        self.last_session = None

    def start(self, caller_frame=None):
        self.last_profile_time = timer()
        self._start_time = time.time()
        if process_time:
            self._start_process_time = process_time()
        if caller_frame is None:
            caller_frame = inspect.currentframe().f_back
        self._start_call_stack = self._call_stack_for_frame(caller_frame)

        setstatprofile(self._profile, self.interval)

    def stop(self):
        setstatprofile(None)
        if process_time:
            cpu_time = process_time() - self._start_process_time
            self._start_process_time = None
        else:
            cpu_time = None

        self.last_session = ProfilerSession(
            frame_records=self.frame_records,
            start_time=self._start_time,
            duration=time.time() - self._start_time,
            sample_count=len(self.frame_records),
            program=' '.join(sys.argv),
            start_call_stack=self._start_call_stack,
            cpu_time=cpu_time,
        )

        return self.last_session

    def __enter__(self):
        self.start(caller_frame=inspect.currentframe().f_back)
        return self

    def __exit__(self, *args):
        self.stop()

    # pylint: disable=W0613
    def _profile(self, frame, event, arg):
        now = timer()
        time_since_last_profile = now - self.last_profile_time

        if event == 'call':
            frame = frame.f_back

        self.frame_records.append((self._call_stack_for_frame(frame), time_since_last_profile))

        self.last_profile_time = now

    def _call_stack_for_frame(self, frame):
        call_stack = []

        while frame is not None:
            identifier = '%s\x00%s\x00%i' % (
                frame.f_code.co_name, frame.f_code.co_filename, frame.f_code.co_firstlineno
            )
            call_stack.append(identifier)
            frame = frame.f_back

        # we iterated from the leaf to the root, we actually want the call stack
        # starting at the root, so reverse this array
        call_stack.reverse()
        return call_stack

    @deprecated_option('root')
    def output_text(self, root=None, unicode=False, color=False, show_all=False, timeline=False):
        return renderers.ConsoleRenderer(unicode=unicode, color=color, show_all=show_all, timeline=timeline).render(self.last_session)

    @deprecated_option('root')
    def output_html(self, root=None):
        return renderers.HTMLRenderer().render(self.last_session)

    def open_in_browser(self):
        return renderers.HTMLRenderer().open_in_browser(self.last_session)

    @deprecated_option('root')
    def output(self, renderer, root=None):
        return renderer.render(self.last_session)

    @deprecated
    def root_frame(self):
        if self.last_session:
            return self.last_session.root_frame()

    @deprecated
    def first_interesting_frame(self):
        """
        Traverse down the frame hierarchy until a frame is found with more than one child
        """
        root_frame = self.root_frame()
        frame = root_frame

        while len(frame.children) <= 1:
            if frame.children:
                frame = frame.children[0]
            else:
                # there are no branches
                return root_frame

        return frame

    @deprecated
    def starting_frame(self, root=False):
        if root:
            return self.root_frame()
        else:
            return self.first_interesting_frame()



import io
import os
import sys
import time

from django.http import HttpResponse
from django.conf import settings
from django.utils.module_loading import import_string
from pyinstrument import Profiler


try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object


class ProfilerMiddleware(MiddlewareMixin):
    def process_request(self, request):
        profile_dir = getattr(settings, 'PYINSTRUMENT_PROFILE_DIR', None)

        func_or_path = getattr(settings, 'PYINSTRUMENT_SHOW_CALLBACK', None)
        if isinstance(func_or_path, str):
            show_pyinstrument = import_string(func_or_path)
        elif callable(func_or_path):
            show_pyinstrument = func_or_path
        else:
            show_pyinstrument = lambda request: True

        if (show_pyinstrument(request) and
            getattr(settings, 'PYINSTRUMENT_URL_ARGUMENT', 'profile') in request.GET
        ) or profile_dir:
            profiler = Profiler()
            profiler.start()

            request.profiler = profiler

    def process_response(self, request, response):
        if hasattr(request, 'profiler'):
            request.profiler.stop()

            output_html = request.profiler.output_html()

            profile_dir = getattr(settings, 'PYINSTRUMENT_PROFILE_DIR', None)

            # Limit the length of the file name (255 characters is the max limit on major current OS, but it is rather
            # high and the other parts (see line 36) are to be taken into account; so a hundred will be fine here).
            path = request.get_full_path().replace('/', '_')[:100]

            # Swap ? for _qs_ on Windows, as it does not support ? in filenames.
            if sys.platform in ['win32', 'cygwin']:
                path = path.replace('?', '_qs_')

            if profile_dir:
                filename = '{total_time:.3f}s {path} {timestamp:.0f}.html'.format(
                    total_time=request.profiler.root_frame().time(),
                    path=path,
                    timestamp=time.time()
                )

                file_path = os.path.join(profile_dir, filename)

                if not os.path.exists(profile_dir):
                    os.mkdir(profile_dir)

                with io.open(file_path, 'w', encoding='utf-8') as f:
                    f.write(output_html)

            if getattr(settings, 'PYINSTRUMENT_URL_ARGUMENT', 'profile') in request.GET:
                return HttpResponse(output_html)
            else:
                return response
        else:
            return response



import setuptools, subprocess, os, distutils
import setuptools.command.build_py
from setuptools import setup, find_packages


HTML_RENDERER_DIR = 'html_renderer'

# pylint: disable=e1101
class CommandUtilities:
    def check_call(self, args, **popen_kwargs):
        self.announce('Running command: %s' % ' '.join(args), level=distutils.log.INFO)
        if not self.dry_run:
            subprocess.check_call(args, **popen_kwargs)


class BuildPyCommand(setuptools.command.build_py.build_py, CommandUtilities):
    """Custom build command."""

    def run(self):
        '''compile the JS, then run superclass implementation'''

        if subprocess.call(['npm', '--version']) != 0:
            raise RuntimeError('npm is required to build the HTML renderer.')

        self.check_call(['npm', 'install'], cwd=HTML_RENDERER_DIR)
        self.check_call(['npm', 'run', 'build'], cwd=HTML_RENDERER_DIR)

        self.copy_file(HTML_RENDERER_DIR+'/dist/js/app.js', 'pyinstrument/renderers/html_resources/app.js')

        setuptools.command.build_py.build_py.run(self)


class HTMLDevServerCommand(distutils.cmd.Command, CommandUtilities):
    description = 'run the HTML renderer dev server'
    user_options = []
    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        subprocess.check_call(['npm', 'run', 'serve'], cwd=HTML_RENDERER_DIR)


class BuildAndUploadCommand(distutils.cmd.Command, CommandUtilities):
    user_options = []
    def initialize_options(self): pass
    def finalize_options(self): pass
    def run(self):
        self.check_call(['rm', '-rf', 'dist'])
        self.run_command('build')
        self.run_command('bdist_wheel')
        self.check_call(['twine upload dist/*'], shell=True)

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    long_description = f.read()

setup(
    name="pyinstrument",
    packages=find_packages(),
    version="3.1.0",
    description="Call stack profiler for Python. Shows you why your code is slow!",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Joe Rickerby',
    author_email='joerick@mac.com',
    url='https://github.com/joerick/pyinstrument',
    keywords=['profiling', 'profile', 'profiler', 'cpu', 'time', 'sampling'],
    install_requires=['pyinstrument_cext>=0.2.2'],
    include_package_data=True,
    entry_points={'console_scripts': ['pyinstrument = pyinstrument.__main__:main']},
    zip_safe=False,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    cmdclass={
        'build_py': BuildPyCommand,
        'dev_server': HTMLDevServerCommand,
        'build_and_upload': BuildAndUploadCommand,
    },
    classifiers=[
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Debuggers',
        'Topic :: Software Development :: Testing',
    ]
)



'''
Processors are functions that take a Frame object, and mutate the tree to perform some task.

They can mutate the tree in-place, but also can change the root frame, they should always be
called like:

    frame = processor(frame, options=...)
'''

import re
from operator import methodcaller
from pyinstrument.frame import FrameGroup, SelfTimeFrame


def remove_importlib(frame, options):
    if frame is None:
        return None

    for child in frame.children:
        remove_importlib(child, options=options)

        if '<frozen importlib._bootstrap' in child.file_path:
            # remove this node, moving the self_time and children up to the parent
            frame.self_time += child.self_time
            frame.add_children(child.children, after=child)
            child.remove_from_parent()

    return frame


def aggregate_repeated_calls(frame, options):
    '''
    Converts a timeline into a time-aggregate summary.

    Adds together calls along the same call stack, so that repeated calls appear as the same
    frame. Removes time-linearity - frames are sorted according to total time spent.

    Useful for outputs that display a summary of execution (e.g. text and html outputs)
    '''
    if frame is None:
        return None

    children_by_identifier = {}

    # iterate over a copy of the children since it's going to mutate while we're iterating
    for child in frame.children:
        if child.identifier in children_by_identifier:
            aggregate_frame = children_by_identifier[child.identifier]

            # combine the two frames, putting the children and self_time into the aggregate frame.
            aggregate_frame.self_time += child.self_time
            if child.children:
                aggregate_frame.add_children(child.children)

            # remove this frame, it's been incorporated into aggregate_frame
            child.remove_from_parent()
        else:
            # never seen this identifier before. It becomes the aggregate frame.
            children_by_identifier[child.identifier] = child

    # recurse into the children
    for child in frame.children:
        aggregate_repeated_calls(child, options=options)

    # sort the children by time
    # it's okay to use the internal _children list, sinde we're not changing the tree
    # structure.
    frame._children.sort(key=methodcaller('time'), reverse=True)  # pylint: disable=W0212

    return frame


def group_library_frames_processor(frame, options):
    if frame is None:
        return None

    hide_regex = options.get('hide_regex', r'.*[\\\/]lib[\\\/].*')
    show_regex = options.get('show_regex')

    def should_be_hidden(frame):
        return ((hide_regex and re.match(hide_regex, frame.file_path))
                and not (show_regex and re.match(show_regex, frame.file_path)))

    def add_frames_to_group(frame, group):
        group.add_frame(frame)
        for child in frame.children:
            if should_be_hidden(child):
                add_frames_to_group(child, group)

    for child in frame.children:
        if not child.group and (should_be_hidden(child)
                                and any(should_be_hidden(cc) for cc in child.children)):
            group = FrameGroup(child)
            add_frames_to_group(child, group)

        group_library_frames_processor(child, options=options)

    return frame    


def merge_consecutive_self_time(frame, options):
    '''
    Combines consecutive 'self time' frames
    '''
    if frame is None:
        return None

    previous_self_time_frame = None

    for child in frame.children:
        if isinstance(child, SelfTimeFrame):
            if previous_self_time_frame:
                # merge
                previous_self_time_frame.self_time += child.self_time
                child.remove_from_parent()
            else:
                # keep a reference, maybe it'll be added to on the next loop
                previous_self_time_frame = child
        else:
            previous_self_time_frame = None
    
    for child in frame.children:
        merge_consecutive_self_time(child, options=options)
    
    return frame


def remove_unnecessary_self_time_nodes(frame, options):
    '''
    When a frame has only one child, and that is a self-time frame, remove that node, since it's
    unnecessary - it clutters the output and offers no additional information.
    '''
    if frame is None:
        return None

    if len(frame.children) == 1 and isinstance(frame.children[0], SelfTimeFrame):
        child = frame.children[0]
        frame.self_time += child.self_time
        child.remove_from_parent()
    
    for child in frame.children:
        remove_unnecessary_self_time_nodes(child, options=options)
    
    return frame


def remove_irrelevant_nodes(frame, options, total_time=None):
    '''
    Remove nodes that represent less than e.g. 1% of the output
    '''
    if frame is None:
        return None
    
    if total_time is None:
        total_time = frame.time()

    filter_threshold = options.get('filter_threshold', 0.01)

    for child in frame.children:
        proportion_of_total = child.time() / total_time

        if proportion_of_total < filter_threshold:
            frame.self_time += child.time()
            child.remove_from_parent()
        
    for child in frame.children:
        remove_irrelevant_nodes(child, options=options, total_time=total_time)

    return frame
